//////////////////////////////////////////////////////////////////////////////////
//     This source code is provided 'as-is', without any express or implied     //
//     warranty. In no event will Infintuary be held liable for any damages     //
//     arising from the use of this software.                                   //
//                                                                              //
//     Infintuary does not warrant, that the source code will be free from      //
//     defects in design or workmanship or that operation of the source code    //
//     will be error-free. No implied or statutory warranty of merchantability  //
//     or fitness for a particular purpose shall apply. The entire risk of      //      
//     quality and performance is with the user of this source code.            //
//                                                                              //
//     Permission is granted to anyone to use this software for any purpose,    //
//     including commercial applications, and to alter it and redistribute it   //
//     freely, subject to the following restrictions:                           //
//                                                                              //
//     1. The origin of this source code must not be misrepresented; you must   //
//        not claim that you wrote the original source code.                    //
//                                                                              //
//     2. Altered source versions must be plainly marked as such, and must not  //
//        be misrepresented as being the original source code.                  //
//                                                                              //
//     3. This notice may not be removed or altered from any source             //
//        distribution.                                                         //
//////////////////////////////////////////////////////////////////////////////////

unit X2H_XML;
{$IFDEF FPC}
{$MODE delphi}
{$ELSE}
{$DEFINE WINDOWS}
{$ENDIF}

interface

uses
  SysUtils, Classes, Graphics, WideStrings;

type
  TStyleType = (stOperator, stAttribute, stQuote, stValue, stComment, stText,
                stCDATA, stProcessing, stDoctype, stTag, stSpace);

  TXML2HTML = class
  private
    Entries: TList;
    FLines: TList;
    FProlog: WideString;
    FBlockIndention: integer;
    FBlockIndentionString: string;

    FFontAttribute: TFont;
    FFontCDATA: TFont;
    FFontComment: TFont;       
    FFontDoctype: TFont;       
    FFontOperator: TFont;      
    FFontProcessing: TFont;    
    FFontQuote: TFont;         
    FFontTag: TFont;           
    FFontText: TFont;          
    FFontValue: TFont;         

    function ParseTokenList(TokenList: TList): boolean;
    procedure BuildLines;
    procedure SetBlockIndention(Value: integer);
    function GetHTMLBlockIndention(Level: integer): string;
    function StyleToString(StyleType: TStyleType): string;
    function ReplaceHTMLChars(ws: WideString): WideString;
    function StyleToHTML(Name: string; Font: TFont): string;
  public
    constructor Create;
    destructor Destroy; override;
    function ParseXML(XML: WideString): boolean;
    function GetAsHTML: WideString;

    property BlockIndention: integer read FBlockIndention write SetBlockIndention;
    property FontAttribute: TFont read FFontAttribute;
    property FontCDATA: TFont read FFontCDATA;
    property FontComment: TFont read FFontComment;
    property FontDoctype: TFont read FFontDoctype;
    property FontOperator: TFont read FFontOperator;
    property FontProcessing: TFont read FFontProcessing;
    property FontQuote: TFont read FFontQuote;
    property FontTag: TFont read FFontTag;
    property FontText: TFont read FFontText;
    property FontValue: TFont read FFontValue;
  end;

  TWordInfo = class
  private
    FText: WideString;
    FStyleType: TStyleType;
  public
    constructor Create(AText: WideString; AStyleType: TStyleType);
    property Text: WideString read FText;
    property StyleType: TStyleType read FStyleType;
  end;

  TLineInfo = class
  private
    FLevel: integer;
    FWordList: TList;
  public
    constructor Create(ALevel: integer);
    destructor Destroy; override;
    procedure Add(WordInfo: TWordInfo);
    property Level: integer read FLevel;
    property WordList: TList read FWordList;
  end;

implementation

{$IFNDEF UNICODE}
type
  TCharSet = set of char;
{$ENDIF}
{$IFNDEF UNICODE}

function CharInSet(c: WideChar; TheSet: TCharSet): boolean;
begin
  result := (char(c) in TheSet);
end;
{$ENDIF}

const
  __DEFAULT_FONT_NAME = 'Courier New';
  __DEFAULT_FONT_SIZE = 12;
  __DEFAULT_FONT_Color = clBlack;

type
  TTokenType = (ttEOF, ttUnknown, ttEntityStart, ttEntityEnd, ttEntityStartEnd,
                ttComment, ttCData, ttProcessing, ttText, ttDoctype);

  TParam = class
    Name: WideString;
    Value: WideString;
    constructor Create(AName, AValue: WideString); overload;
    constructor Create(AParam: TParam); overload;
  end;

  TEntry = class
    procedure BuildLines(Lines: TList; Level: integer); virtual; abstract;
  end;

  TText = class(TEntry)
    Text: TWideStringList;
    constructor Create(AText: WideString);
    destructor Destroy; override;
    function SingleLine: boolean;
    procedure BuildLines(Lines: TList; Level: integer); override;
  end;

  TEmptyTag = class(TEntry)
    Name: WideString;
    Params: TList;
    constructor Create(AName: WideString; AParams: TList);
    destructor Destroy; override;
    procedure BuildLines(Lines: TList; Level: integer); override;
  end;

  TComment = class(TEntry) // <!--
    Comment: WideString;
    constructor Create(AComment: WideString);
    procedure BuildLines(Lines: TList; Level: integer); override;
  end;

  TCData = class(TEntry) // <!CDATA[
    Data: WideString;
    constructor Create(AData: WideString);
    procedure BuildLines(Lines: TList; Level: integer); override;
  end;

  TProcessing = class(TEntry) // <?
    Name: WideString;
    Params: TList;
    constructor Create(AName: WideString; AParams: TList);
    destructor Destroy; override;
    procedure BuildLines(Lines: TList; Level: integer); override;
    function GetAsString: WideString;
  end;

  TDoctype = class(TEntry) // <!DOCTYPE
    Name: WideString;
    Params: WideString;
    constructor Create(AName: WideString; AParams: WideString);
    procedure BuildLines(Lines: TList; Level: integer); override;
  end;

  TEntity = class(TEntry)
    Name: WideString;
    Params: TList;
    Entries: TList;
    constructor Create(AName: WideString; AParams: TList);
    destructor Destroy; override;
    function ParseTokenList(TokenList: TList; var Index: integer): boolean;
    procedure BuildLines(Lines: TList; Level: integer); override;
  end;

  TToken = class
    Value: WideString;
    Params: WideString;
    TokenType: TTokenType;
    ParamList: TList;
    constructor Create(ATokenType: TTokenType);
    destructor Destroy; override;
    procedure AddParam(AttName, AttValue: WideString);
  end;

constructor TWordInfo.Create(AText: WideString; AStyleType: TStyleType);
begin
  FText := AText;
  FStyleType := AStyleType;
end;

constructor TLineInfo.Create(ALevel: integer);
begin
  FLevel := ALevel;
  FWordList := TList.Create;
end;

destructor TLineInfo.Destroy;
var
  i: integer;
begin
  for i := 0 to FWordList.Count - 1 do TWordInfo(FWordList[i]).Free;
  FWordList.Free;
  inherited;
end;

procedure TLineInfo.Add(WordInfo: TWordInfo);
begin
  FWordList.Add(WordInfo);
end;

constructor TToken.Create(ATokenType: TTokenType);
begin
  Value := '';
  Params := '';
  TokenType := ATokenType;
  ParamList := TList.Create;
end;

destructor TToken.Destroy;
var
  i: integer;
begin
  for i := 0 to ParamList.Count - 1 do TParam(ParamList[i]).Free;
  ParamList.Free;
  inherited;
end;

procedure TToken.AddParam(AttName, AttValue: WideString);
begin
  if (AttName <> '') and (AttValue <> '') then ParamList.Add(TParam.Create(AttName, AttValue));
end;

constructor TParam.Create(AName, AValue: WideString);
begin
  inherited Create;
  Name := AName;
  Value := AValue;
end;

constructor TParam.Create(AParam: TParam);
begin
  inherited Create;
  Name := AParam.Name;
  Value := AParam.Value;
end;

constructor TText.Create(AText: WideString);
begin
  inherited Create;
  Text := TWideStringList.Create;
  Text.Text := trim(AText);
end;

destructor TText.Destroy;
begin
  Text.Free;
  inherited;
end;

function TText.SingleLine: boolean;
begin
  result := Text.Count < 2;
end;

procedure TText.BuildLines(Lines: TList; Level: integer);
var
  i: integer;
  LineInfo: TLineInfo;
begin
  for i := 0 to Text.Count - 1 do begin
    LineInfo := TLineInfo.Create(Level);
    Lines.Add(LineInfo);
    LineInfo.Add(TWordInfo.Create(Text[i], stText));
  end;
end;

constructor TEmptyTag.Create(AName: WideString; AParams: TList);
var
  i: integer;
begin
  Name := AName;
  Params := TList.Create;
  for i := 0 to AParams.Count - 1 do begin
    Params.Add(TParam.Create(TParam(AParams[i])));
  end;
end;

destructor TEmptyTag.Destroy;
var
  i: integer;
begin
  for i := 0 to Params.Count - 1 do TParam(Params[i]).Free;
  Params.Free;
end;

procedure TEmptyTag.BuildLines(Lines: TList; Level: integer);
var
  i: integer;
  Param: TParam;
  LineInfo: TLineInfo;
begin
  LineInfo := TLineInfo.Create(Level);
  Lines.Add(LineInfo);
  LineInfo.Add(TWordInfo.Create('<', stOperator));
  LineInfo.Add(TWordInfo.Create(Name, stTag));

  for i := 0 to Params.Count - 1 do begin
    Param := TParam(Params[i]);
    LineInfo.Add(TWordInfo.Create(' ', stSpace));
    LineInfo.Add(TWordInfo.Create(Param.Name, stAttribute));
    LineInfo.Add(TWordInfo.Create('=', stOperator));
    LineInfo.Add(TWordInfo.Create('"', stQuote));
    LineInfo.Add(TWordInfo.Create(Param.Value, stValue));
    LineInfo.Add(TWordInfo.Create('"', stQuote));
  end;
  LineInfo.Add(TWordInfo.Create('/>', stOperator));
end;

constructor TProcessing.Create(AName: WideString; AParams: TList);
var
  i: integer;
begin
  inherited Create;
  Name := AName;
  Params := TList.Create;
  for i := 0 to AParams.Count - 1 do begin
    Params.Add(TParam.Create(TParam(AParams[i])));
  end;
end;

destructor TProcessing.Destroy;
var
  i: integer;
begin
  for i := 0 to Params.Count - 1 do TParam(Params[i]).Free;
  inherited;
end;

procedure TProcessing.BuildLines(Lines: TList; Level: integer);
var
  i: integer;
  Param: TParam;
  LineInfo: TLineInfo;
begin
  LineInfo := TLineInfo.Create(Level);
  Lines.Add(LineInfo);
  LineInfo.Add(TWordInfo.Create('<?', stOperator));
  LineInfo.Add(TWordInfo.Create(Name, stProcessing));

  for i := 0 to Params.Count - 1 do begin
    Param := TParam(Params[i]);
    LineInfo.Add(TWordInfo.Create(' ', stSpace));
    LineInfo.Add(TWordInfo.Create(Param.Name, stAttribute));
    LineInfo.Add(TWordInfo.Create('=', stOperator));
    LineInfo.Add(TWordInfo.Create('"', stQuote));
    LineInfo.Add(TWordInfo.Create(Param.Value, stValue));
    LineInfo.Add(TWordInfo.Create('"', stQuote));
  end;

  LineInfo.Add(TWordInfo.Create(' ?>', stOperator));
end;

function TProcessing.GetAsString: WideString;
var
  i: integer;
  Param: TParam;
begin
  result := '<?' + Name;
  for i := 0 to Params.Count - 1 do begin
    Param := TParam(Params[i]);
    result := result + ' ' + Param.Name + '="' + Param.Value + '"';
  end;
  result := result + ' ?>';
end;

constructor TDoctype.Create(AName: WideString; AParams: WideString);
begin
  Name := AName;
  Params := AParams;
end;

procedure TDoctype.BuildLines(Lines: TList; Level: integer);
var
  LineInfo: TLineInfo;
begin
  LineInfo := TLineInfo.Create(Level);
  Lines.Add(LineInfo);
  LineInfo.Add(TWordInfo.Create('<!', stOperator));
  LineInfo.Add(TWordInfo.Create(Name, stDoctype));
  LineInfo.Add(TWordInfo.Create(' ', stSpace));
  LineInfo.Add(TWordInfo.Create(Params, stDoctype));
  LineInfo.Add(TWordInfo.Create('>', stOperator));
end;

constructor TEntity.Create(AName: WideString; AParams: TList);
var
  i: integer;
begin
  inherited Create;
  Name := AName;
  Params := TList.Create;
  Entries := TList.Create;
  for i := 0 to AParams.Count - 1 do begin
    Params.Add(TParam.Create(TParam(AParams[i])));
  end;
end;

destructor TEntity.Destroy;
var
  i: integer;
begin
  for i := 0 to Entries.Count - 1 do TEntry(Entries[i]).Free;
  Entries.Free;
  for i := 0 to Params.Count - 1 do TParam(Params[i]).Free;
  Params.Free;
  inherited;
end;

function TEntity.ParseTokenList(TokenList: TList; var Index: integer): boolean;
var
  Token: TToken;
  Entity: TEntity;
begin
  result := false;
  if TokenList = nil then exit;
  while Index < TokenList.Count do begin
    Token := TToken(TokenList[Index]);
    case Token.TokenType of
      ttEntityStart:
        begin
          Entity := TEntity.Create(Token.Value, Token.ParamList);
          Entries.Add(Entity);
          inc(Index);
          if not Entity.ParseTokenList(TokenList, Index) then exit;
        end;
      ttEntityEnd:
        begin
          result := Token.Value = Name;
          exit;
        end;
      ttEntityStartEnd: Entries.Add(TEmptyTag.Create(Token.Value, Token.ParamList));
      ttComment: Entries.Add(TComment.Create(Token.Value));
      ttCData: Entries.Add(TCData.Create(Token.Value));
      ttProcessing: Entries.Add(TProcessing.Create(Token.Value, Token.ParamList));
      ttText: Entries.Add(TText.Create(Token.Value));
      ttDoctype: exit;
    end;
    inc(Index);
  end;
end;

procedure TEntity.BuildLines(Lines: TList; Level: integer);
var
  i: integer;
  Param: TParam;
  SingleLine: boolean;
  LineInfo: TLineInfo;
begin
  LineInfo := TLineInfo.Create(Level);
  Lines.Add(LineInfo);
  LineInfo.Add(TWordInfo.Create('<', stOperator));
  LineInfo.Add(TWordInfo.Create(Name, stTag));

  for i := 0 to Params.Count - 1 do begin
    Param := TParam(Params[i]);
    LineInfo.Add(TWordInfo.Create(' ', stSpace));
    LineInfo.Add(TWordInfo.Create(Param.Name, stAttribute));
    LineInfo.Add(TWordInfo.Create('=', stOperator));
    LineInfo.Add(TWordInfo.Create('"', stQuote));
    LineInfo.Add(TWordInfo.Create(Param.Value, stValue));
    LineInfo.Add(TWordInfo.Create('"', stQuote));
  end;

  if Entries.Count = 0 then begin
    LineInfo.Add(TWordInfo.Create('/>', stOperator));
  end
  else begin
    if (Entries.Count = 1) and (TEntry(Entries[0]) is TText) then begin
      SingleLine := TText(Entries[0]).SingleLine;
    end
    else SingleLine := false;

    if SingleLine then begin
      LineInfo.Add(TWordInfo.Create('>', stOperator));
      LineInfo.Add(TWordInfo.Create(TText(Entries[0]).Text[0], stText));

      LineInfo.Add(TWordInfo.Create('</', stOperator));
      LineInfo.Add(TWordInfo.Create(Name, stTag));
      LineInfo.Add(TWordInfo.Create('>', stOperator));
    end
    else begin
      LineInfo.Add(TWordInfo.Create('>', stOperator));
      for i := 0 to Entries.Count - 1 do begin
        TEntry(Entries[i]).BuildLines(Lines, Level + 1);
      end;
      LineInfo := TLineInfo.Create(Level);
      Lines.Add(LineInfo);
      LineInfo.Add(TWordInfo.Create('</', stOperator));
      LineInfo.Add(TWordInfo.Create(Name, stTag));
      LineInfo.Add(TWordInfo.Create('>', stOperator));
    end;
  end;
end;

constructor TComment.Create(AComment: WideString);
begin
  inherited Create;
  Comment := AComment;
end;

procedure TComment.BuildLines(Lines: TList; Level: integer);
var
  i: integer;
  LineInfo: TLineInfo;
  tmpSL: TWideStringList;
begin
  tmpSL := TWideStringList.Create;
  try
    tmpSL.Text := Comment;

    LineInfo := TLineInfo.Create(Level);
    Lines.Add(LineInfo);
    LineInfo.Add(TWordInfo.Create('<!--', stOperator));

    if tmpSL.Count = 1 then begin
      LineInfo.Add(TWordInfo.Create(Comment, stComment));
    end
    else begin
      for i := 0 to tmpSL.Count - 1 do begin
        LineInfo := TLineInfo.Create(Level + 1);
        Lines.Add(LineInfo);
        LineInfo.Add(TWordInfo.Create(tmpSL[i], stComment));
      end;
      LineInfo := TLineInfo.Create(Level);
      Lines.Add(LineInfo);
    end;
    LineInfo.Add(TWordInfo.Create('-->', stOperator));
  finally
    tmpSL.Free;
  end;
end;

constructor TCData.Create(AData: WideString);
begin
  inherited Create;
  Data := AData;
end;

procedure TCData.BuildLines(Lines: TList; Level: integer);
var
  i: integer;
  LineInfo: TLineInfo;
  tmpSL: TWideStringList;
begin
  tmpSL := TWideStringList.Create;
  try
    tmpSL.Text := trim(Data);

    LineInfo := TLineInfo.Create(Level);
    Lines.Add(LineInfo);
    LineInfo.Add(TWordInfo.Create('<!CDATA[', stOperator));

    if tmpSL.Count = 1 then begin
      LineInfo.Add(TWordInfo.Create(tmpSL[0], stCDATA));
    end
    else begin
      for i := 0 to tmpSL.Count - 1 do begin
        LineInfo := TLineInfo.Create(Level + 1);
        Lines.Add(LineInfo);
        LineInfo.Add(TWordInfo.Create(trim(tmpSL[i]), stCDATA));
      end;
      LineInfo := TLineInfo.Create(Level);
      Lines.Add(LineInfo);
    end;
    LineInfo.Add(TWordInfo.Create(']]>', stOperator));
  finally
    tmpSL.Free;
  end;
end;

constructor TXML2HTML.Create;
begin
  inherited;
  Entries := TList.Create;
  FLines := TList.Create;
  FProlog := '';
  BlockIndention := 4;

  FFontAttribute := TFont.Create;
  FFontCDATA := TFont.Create;
  FFontComment := TFont.Create;
  FFontDoctype := TFont.Create;
  FFontOperator := TFont.Create;
  FFontProcessing := TFont.Create;
  FFontQuote := TFont.Create;
  FFontTag := TFont.Create;
  FFontText := TFont.Create;
  FFontValue := TFont.Create;

  FFontText.Name := __DEFAULT_FONT_NAME;
  FFontText.Size := __DEFAULT_FONT_SIZE;
  FFontText.Color := __DEFAULT_FONT_Color;

  FFontAttribute.Assign(FFontText);
  FFontCDATA.Assign(FFontText);
  FFontComment.Assign(FFontText);
  FFontDoctype.Assign(FFontText);
  FFontOperator.Assign(FFontText);
  FFontProcessing.Assign(FFontText);
  FFontQuote.Assign(FFontText);
  FFontTag.Assign(FFontText);
  FFontValue.Assign(FFontText);

  FFontAttribute.Color := clRed;
  FFontCDATA.Color := clPurple;
  FFontComment.Color := clGray;
  FFontDoctype.Color := clBlue;
  FFontOperator.Color := clBlue;
  FFontProcessing.Color := clBlue;
  FFontTag.Color := clMaroon;
  FFontValue.Color := clGreen;

  FFontComment.Style := [fsItalic];
  FFontText.Style := [fsBold];
  FFontValue.Style := [fsBold];
end;

destructor TXML2HTML.Destroy;
var
  i: integer;
begin
  for i := 0 to Entries.Count - 1 do TEntry(Entries[i]).Free;
  Entries.Free;

  for i := 0 to FLines.Count - 1 do TLineInfo(FLines[i]).Free;
  FLines.Free;

  FFontAttribute.Free;
  FFontCDATA.Free;
  FFontComment.Free;
  FFontDoctype.Free;
  FFontOperator.Free;
  FFontProcessing.Free;
  FFontQuote.Free;
  FFontTag.Free;
  FFontText.Free;
  FFontValue.Free;
  inherited;
end;

function GetToken(Line: WideString; Token: TToken; var p: integer): boolean;
const
  ST_START               = 0;
  ST_START2              = 1;
  ST_TAG_START           = 2;
  ST_START_TAG_NAME      = 3;
  ST_START_TAG_NAME_END  = 4;
  ST_START_TAG_NAME_END2 = 5;
  ST_START_TAG_SPACE     = 6;
  ST_ATT_NAME            = 7;
  ST_ATT_NAME_SPACE      = 8;
  ST_ATT_NAME_EQUAL      = 9;
  ST_ATT_VALUE_START1    = 10;
  ST_ATT_VALUE_START2    = 11;
  ST_ATT_VALUE_END       = 12;
  ST_START_TAG_END       = 13;
  ST_END_TAG_START1      = 14;
  ST_END_TAG             = 15;
  ST_END_TAG_SPACE       = 16;
  ST_PROCESSING_START    = 17;
  ST_PROCESSING_NAME     = 18;
  ST_PROCESSING_SPACE    = 19;
  ST_PROCESSING_END      = 20;
  ST_COMMENT_START1      = 21;
  ST_COMMENT_START2      = 22;
  ST_COMMENT             = 23;
  ST_COMMENT_END1        = 24;
  ST_COMMENT_END2        = 25;
  ST_CDATA_START1        = 26;
  ST_CDATA_START2        = 27;
  ST_CDATA_START3        = 28;
  ST_CDATA_START4        = 29;
  ST_CDATA_START5        = 30;
  ST_CDATA_START6        = 31;
  ST_CDATA               = 32;
  ST_CDATA_END1          = 33;
  ST_CDATA_END2          = 34;
  ST_PATT_NAME           = 35;
  ST_PATT_NAME_SPACE     = 36;
  ST_PATT_NAME_EQUAL     = 37;
  ST_PATT_VALUE_START1   = 38;
  ST_PATT_VALUE_START2   = 39;
  ST_PATT_VALUE_END      = 40;
  ST_DOCTYPE_START       = 41;
  ST_DOCTYPE             = 42;
  ST_DOCTYPE_VALUE1      = 43;
  ST_DOCTYPE_VALUE2      = 44;
  ST_DOCTYPE_VALUE3      = 45;

  Point        = '.';
  Colon        = ':';
  Hyphen       = '.';
  Digits       = ['0' .. '9'];
  Letters      = ['a' .. 'z', 'A' .. 'Z', '_'];
  IDStartChars = Letters;
  IDChars      = Letters + Digits + [Colon] + [Point] + [Hyphen];
  Space        = [#32, #9, #10, #13];
  Equal        = '=';
  EndOfFile    = #0;
var
  c: WideChar;
  State: integer;
  StartP: integer;
  AttName: WideString;
  AttValue: WideString;
begin
  result := false;

  StartP := p;
  State := ST_START;

  while true do begin
    if p <= length(Line) then c := Line[p]
    else c := EndOfFile;
    case State of
      ST_START:
        begin
          if c = '<' then begin
            if StartP = p then State := ST_TAG_START
            else begin
              Token.TokenType := ttText;
              Token.Value := copy(Line, StartP, p - StartP);
              result := true;
              exit;
            end;
          end
          else if c = EndOfFile then begin
            if StartP = p then Token.TokenType := ttEOF
            else begin
              Token.TokenType := ttText;
              Token.Value := copy(Line, StartP, p - StartP);
            end;
            result := true;
            exit;
          end
          else if CharInSet(c, Space) then begin
            StartP := p + 1;
          end
          else State := ST_START2;
        end;
      ST_START2:
        begin
          if c = '<' then begin
            if StartP = p then State := ST_TAG_START
            else begin
              Token.TokenType := ttText;
              Token.Value := trim(copy(Line, StartP, p - StartP));
              result := true;
              exit;
            end;
          end
          else if c = EndOfFile then begin
            if StartP = p then Token.TokenType := ttEOF
            else begin
              Token.TokenType := ttText;
              Token.Value := trim(copy(Line, StartP, p - StartP));
            end;
            result := true;
            exit;
          end
          else State := ST_START2;
        end;

      ST_TAG_START: if c = '!' then State := ST_COMMENT_START1
        else if c = '?' then State := ST_PROCESSING_START
        else if c = '/' then State := ST_END_TAG_START1
        else if CharInSet(c, IDStartChars) then State := ST_START_TAG_NAME
        else exit;

      ST_START_TAG_NAME: if CharInSet(c, IDChars) then State := ST_START_TAG_NAME
        else if CharInSet(c, Space) then begin
          Token.Value := trim(copy(Line, StartP + 1, p - StartP));
          StartP := p + 1;
          State := ST_START_TAG_SPACE;
        end
        else if c = '>' then begin
          inc(p);
          Token.TokenType := ttEntityStart;
          Token.Value := trim(copy(Line, StartP + 1, p - StartP - 2));
          result := true;
          exit;
        end
        else if c = '/' then State := ST_START_TAG_NAME_END
        else exit;
      ST_START_TAG_NAME_END: if c = '>' then begin
          inc(p);
          Token.TokenType := ttEntityStartEnd;
          Token.Value := trim(copy(Line, StartP + 1, p - StartP - 3));
          result := true;
          exit;
        end
        else exit;
      ST_START_TAG_NAME_END2: if c = '>' then begin
          inc(p);
          Token.TokenType := ttEntityStartEnd;
          result := true;
          exit;
        end
        else exit;
      ST_START_TAG_SPACE: if CharInSet(c, Space) then State := ST_START_TAG_SPACE
        else if c = '/' then State := ST_START_TAG_NAME_END2
        else if CharInSet(c, IDStartChars) then begin
          State := ST_ATT_NAME;
          StartP := p;
          AttName := '';
          AttValue := '';
        end
        else if c = '>' then begin
          inc(p);
          Token.TokenType := ttEntityStart;
          result := true;
          exit;
        end
        else exit;
      ST_ATT_NAME: if CharInSet(c, IDChars) then State := ST_ATT_NAME
        else if CharInSet(c, Space) then State := ST_ATT_NAME_SPACE
        else if c = Equal then begin
          AttName := trim(copy(Line, StartP, p - StartP));
          State := ST_ATT_NAME_EQUAL;
        end
        else exit;
      ST_ATT_NAME_SPACE: if CharInSet(c, Space) then State := ST_ATT_NAME_SPACE
        else if c = Equal then begin
          AttName := trim(copy(Line, StartP, p - StartP));
          State := ST_ATT_NAME_EQUAL;
        end
        else exit;
      ST_ATT_NAME_EQUAL: if CharInSet(c, Space) then State := ST_ATT_NAME_EQUAL
        else if c = '''' then begin
          StartP := p;
          State := ST_ATT_VALUE_START1;
        end
        else if c = '"' then begin
          StartP := p;
          State := ST_ATT_VALUE_START2;
        end
        else exit;
      ST_ATT_VALUE_START1: if c = '''' then begin
          AttValue := trim(copy(Line, StartP + 1, p - StartP - 1));
          Token.AddParam(AttName, AttValue);
          State := ST_ATT_VALUE_END;
        end
        else if c = EndOfFile then exit
        else State := ST_ATT_VALUE_START1;
      ST_ATT_VALUE_START2: if c = '"' then begin
          AttValue := trim(copy(Line, StartP + 1, p - StartP - 1));
          Token.AddParam(AttName, AttValue);
          State := ST_ATT_VALUE_END;
        end
        else if c = EndOfFile then exit
        else State := ST_ATT_VALUE_START2;
      ST_ATT_VALUE_END: if CharInSet(c, Space) then State := ST_ATT_VALUE_END
        else if CharInSet(c, IDStartChars) then begin
          State := ST_ATT_NAME;
          StartP := p;
        end
        else if c = '/' then State := ST_START_TAG_END
        else if c = '>' then begin
          inc(p);
          Token.TokenType := ttEntityStart;
          Token.Params := trim(copy(Line, StartP, p - StartP - 1));
          result := true;
          exit;
        end
        else exit;
      ST_START_TAG_END: if c = '>' then begin
          inc(p);
          Token.TokenType := ttEntityStartEnd;
          Token.Params := trim(copy(Line, StartP + 1, p - StartP - 3));
          result := true;
          exit;
        end
        else exit;

      ST_END_TAG_START1: if CharInSet(c, IDStartChars) then State := ST_END_TAG
        else exit;
      ST_END_TAG: if CharInSet(c, IDChars) then State := ST_END_TAG
        else if CharInSet(c, Space) then State := ST_END_TAG_SPACE
        else if c = '>' then begin
          inc(p);
          Token.TokenType := ttEntityEnd;
          Token.Value := trim(copy(Line, StartP + 2, p - StartP - 3));
          result := true;
          exit;
        end
        else exit;
      ST_END_TAG_SPACE: if CharInSet(c, Space) then State := ST_END_TAG_SPACE
        else if c = '>' then begin
          inc(p);
          Token.TokenType := ttEntityEnd;
          Token.Value := trim(copy(Line, StartP + 2, p - StartP - 3));
          result := true;
          exit;
        end
        else exit;

      ST_PROCESSING_START: if CharInSet(c, IDStartChars) then State := ST_PROCESSING_NAME
        else exit;
      ST_PROCESSING_NAME: if CharInSet(c, IDChars) then State := ST_PROCESSING_NAME
        else if CharInSet(c, Space) then begin
          Token.Value := trim(copy(Line, StartP + 2, p - StartP - 1));
          StartP := p;
          State := ST_PROCESSING_SPACE;
        end
        else if c = '?' then begin
          Token.Value := trim(copy(Line, StartP + 2, p - StartP - 1));
          StartP := p;
          State := ST_PROCESSING_END;
        end
        else exit;
      ST_PROCESSING_END: if c = '>' then begin
          inc(p);
          Token.TokenType := ttProcessing;
          result := true;
          exit;
        end
        else exit;
      ST_PROCESSING_SPACE: if CharInSet(c, Space) then State := ST_PROCESSING_SPACE
        else if CharInSet(c, IDStartChars) then begin
          State := ST_PATT_NAME;
          StartP := p;
          AttName := '';
          AttValue := '';
        end
        else if c = '?' then State := ST_PROCESSING_END
        else exit;
      ST_PATT_NAME: if CharInSet(c, IDChars) then State := ST_PATT_NAME
        else if CharInSet(c, Space) then State := ST_PATT_NAME_SPACE
        else if c = Equal then begin
          AttName := trim(copy(Line, StartP, p - StartP));
          State := ST_PATT_NAME_EQUAL;
        end
        else exit;
      ST_PATT_NAME_SPACE: if CharInSet(c, Space) then State := ST_PATT_NAME_SPACE
        else if c = Equal then begin
          AttName := trim(copy(Line, StartP, p - StartP));
          State := ST_PATT_NAME_EQUAL;
        end
        else exit;
      ST_PATT_NAME_EQUAL: if CharInSet(c, Space) then State := ST_PATT_NAME_EQUAL
        else if c = '''' then begin
          StartP := p;
          State := ST_PATT_VALUE_START1;
        end
        else if c = '"' then begin
          StartP := p;
          State := ST_PATT_VALUE_START2;
        end
        else exit;
      ST_PATT_VALUE_START1: if c = '''' then begin
          AttValue := trim(copy(Line, StartP + 1, p - StartP - 1));
          Token.AddParam(AttName, AttValue);
          State := ST_PATT_VALUE_END;
        end
        else if c = EndOfFile then exit
        else State := ST_PATT_VALUE_START1;
      ST_PATT_VALUE_START2: if c = '"' then begin
          AttValue := trim(copy(Line, StartP + 1, p - StartP - 1));
          Token.AddParam(AttName, AttValue);
          State := ST_PATT_VALUE_END;
        end
        else if c = EndOfFile then exit
        else State := ST_PATT_VALUE_START2;
      ST_PATT_VALUE_END: if CharInSet(c, Space) then State := ST_PATT_VALUE_END
        else if CharInSet(c, IDStartChars) then begin
          State := ST_PATT_NAME;
          StartP := p;
        end
        else if c = '?' then State := ST_PROCESSING_END
        else exit;

      ST_DOCTYPE_START: if CharInSet(c, IDChars) then State := ST_DOCTYPE_START
        else if CharInSet(c, Space) then begin
          Token.Value := trim(copy(Line, StartP + 2, p - StartP - 1));
          if Token.Value <> 'DOCTYPE' then exit;
          StartP := p;
          State := ST_DOCTYPE;
        end
        else exit;
      ST_DOCTYPE: if CharInSet(c, Space) then State := ST_DOCTYPE
        else if CharInSet(c, IDChars) then State := ST_DOCTYPE
        else if c = '"' then State := ST_DOCTYPE_VALUE1
        else if c = '''' then State := ST_DOCTYPE_VALUE2
        else if c = '[' then State := ST_DOCTYPE_VALUE3
        else if c = '>' then begin
          inc(p);
          Token.Params := trim(copy(Line, StartP + 1, p - StartP - 2));
          Token.TokenType := ttDoctype;
          result := true;
          exit;
        end
        else exit;
      ST_DOCTYPE_VALUE1: if c = '"' then State := ST_DOCTYPE
        else if c = EndOfFile then exit
        else State := ST_DOCTYPE_VALUE1;
      ST_DOCTYPE_VALUE2: if c = '''' then State := ST_DOCTYPE
        else if c = EndOfFile then exit
        else State := ST_DOCTYPE_VALUE2;
      ST_DOCTYPE_VALUE3: if c = ']' then State := ST_DOCTYPE
        else if c = EndOfFile then exit
        else State := ST_DOCTYPE_VALUE3;

      ST_COMMENT_START1: if c = '-' then State := ST_COMMENT_START2
        else if c = '[' then State := ST_CDATA_START1
        else if c = 'D' then State := ST_DOCTYPE_START
        else exit;
      ST_COMMENT_START2: if c = '-' then begin
          StartP := p + 1;
          State := ST_COMMENT;
        end
        else exit;
      ST_COMMENT: if c = '-' then State := ST_COMMENT_END1
        else if c = EndOfFile then exit
        else State := ST_COMMENT;
      ST_COMMENT_END1: if c = '-' then State := ST_COMMENT_END2
        else if c = EndOfFile then exit
        else State := ST_COMMENT;
      ST_COMMENT_END2: if c = '>' then begin
          inc(p);
          Token.TokenType := ttComment;
          Token.Value := copy(Line, StartP, p - StartP - 3);
          result := true;
          exit;
        end
        else if c = EndOfFile then exit
        else State := ST_COMMENT;

      ST_CDATA_START1: if c = 'C' then State := ST_CDATA_START2
        else exit;
      ST_CDATA_START2: if c = 'D' then State := ST_CDATA_START3
        else exit;
      ST_CDATA_START3: if c = 'A' then State := ST_CDATA_START4
        else exit;
      ST_CDATA_START4: if c = 'T' then State := ST_CDATA_START5
        else exit;
      ST_CDATA_START5: if c = 'A' then State := ST_CDATA_START6
        else exit;
      ST_CDATA_START6: if c = '[' then begin
          StartP := p + 1;
          State := ST_CDATA;
        end
        else exit;
      ST_CDATA: if c = ']' then State := ST_CDATA_END1
        else if c = EndOfFile then exit
        else State := ST_CDATA;
      ST_CDATA_END1: if c = ']' then State := ST_CDATA_END2
        else if c = EndOfFile then exit
        else State := ST_CDATA;
      ST_CDATA_END2: if c = '>' then begin
          inc(p);
          Token.TokenType := ttCData;
          Token.Value := copy(Line, StartP, p - StartP - 3);
          result := true;
          exit;
        end
        else if c = EndOfFile then exit
        else State := ST_CDATA;
    end;
    if (p <= length(Line)) then inc(p);
  end;
end;

function ParseXMLIntoToken(XML: WideString; TokenList: TList): boolean;
var
  p: integer;
  Token: TToken;
begin
  result := false;
  if TokenList = nil then exit;
  for p := 0 to TokenList.Count - 1 do TToken(TokenList[p]).Free;

  p := 1;
  while true do begin
    Token := TToken.Create(ttUnknown);
    if not GetToken(XML, Token, p) then begin
      Token.Free;
      break;
    end
    else if (Token.TokenType = ttEOF) then begin
      Token.Free;
      result := true;
      break;
    end
    else if (Token.TokenType = ttUnknown) then begin
      Token.Free;
      result := false;
      break;
    end
    else TokenList.Add(Token);
  end;
end;

function TXML2HTML.ParseTokenList(TokenList: TList): boolean;
var
  Token: TToken;
  Index: integer;
  Entity: TEntity;
  EntityFound: boolean;
  Processing: TProcessing;
begin
  result := false;
  if TokenList = nil then exit;

  Index := 0;
  if Index >= TokenList.Count then exit;
  Token := TToken(TokenList[Index]);
  if Token.TokenType = ttProcessing then begin
    if lowercase(Token.Value) <> 'xml' then exit;
    Processing := TProcessing.Create(Token.Value, Token.ParamList);
    FProlog := Processing.GetAsString;
    Entries.Add(Processing);
    inc(Index);
  end;

  if Index >= TokenList.Count then exit;
  Token := TToken(TokenList[Index]);
  if Token.TokenType = ttDoctype then begin
    if Token.Value <> 'DOCTYPE' then exit;
    Entries.Add(TDoctype.Create(Token.Value, Token.Params));
    inc(Index);
  end;

  EntityFound := false;
  while Index < TokenList.Count do begin
    Token := TToken(TokenList[Index]);
    case Token.TokenType of
      ttEntityStart:
        begin
          if EntityFound then exit;
          Entity := TEntity.Create(Token.Value, Token.ParamList);
          Entries.Add(Entity);
          inc(Index);
          if not Entity.ParseTokenList(TokenList, Index) then exit;
          EntityFound := true;
        end;
      ttEntityStartEnd:
        begin
          if EntityFound then exit;
          Entries.Add(TEmptyTag.Create(Token.Value, Token.ParamList));
          EntityFound := true;
        end;
      ttEntityEnd: exit;
      ttComment: Entries.Add(TComment.Create(Token.Value));
      ttCData: Entries.Add(TCData.Create(Token.Value));
      ttProcessing: Entries.Add(TProcessing.Create(Token.Value, Token.ParamList));
      ttText: Entries.Add(TText.Create(Token.Value));
      ttDoctype: exit;
    end;
    inc(Index);
  end;
  result := EntityFound;
end;

function TXML2HTML.ParseXML(XML: WideString): boolean;
var
  i: integer;
  TokenList: TList;
begin
  result := false;
  TokenList := TList.Create;
  try
    FProlog := '';
    if not ParseXMLIntoToken(XML, TokenList) then exit;
    if not ParseTokenList(TokenList) then exit;
    BuildLines;
    result := true;
  finally
    for i := 0 to TokenList.Count - 1 do TToken(TokenList[i]).Free;
    TokenList.Free;
  end;
end;

procedure TXML2HTML.BuildLines;
var
  i: integer;
  Entry: TEntry;
  Level: integer;
begin
  for i := 0 to FLines.Count - 1 do TLineInfo(FLines[i]).Free;
  FLines.Clear;

  Level := 0;
  for i := 0 to Entries.Count - 1 do begin
    Entry := TEntry(Entries[i]);
    Entry.BuildLines(FLines, Level);
  end;
end;

procedure TXML2HTML.SetBlockIndention(Value: integer);
var
  i: integer;
begin
  FBlockIndention := Value;
  FBlockIndentionString := '';
  if FBlockIndention > 16 then FBlockIndention := 16;
  for i := 1 to FBlockIndention do FBlockIndentionString := FBlockIndentionString + '&nbsp;';
end;

function TXML2HTML.GetHTMLBlockIndention(Level: integer): string;
var
  i: integer;
begin
  result := '';
  for i := 1 to Level do result := result + FBlockIndentionString;
end;

function TXML2HTML.StyleToString(StyleType: TStyleType): string;
begin
  case StyleType of
    stOperator: result := 'op';
    stAttribute: result := 'att';
    stQuote: result := 'qut';
    stValue: result := 'val';
    stComment: result := 'com';
    stCDATA: result := 'cdata';
    stProcessing: result := 'proc';
    stDoctype: result := 'doc';
    stTag: result := 'tag';
  else result := 'txt';
  end;
end;

function TXML2HTML.ReplaceHTMLChars(ws: WideString): WideString;
begin
  ws := StringReplace(ws, '<', '&lt;', [rfReplaceAll]);
  ws := StringReplace(ws, '>', '&gt;', [rfReplaceAll]);
  result := ws;
end;

function Color2HTML(AColor: TColor): string;
begin
  result := '#' + IntToHex(AColor and $FF, 2) + IntToHex((AColor shr 8) and $FF, 2) + IntToHex((AColor shr 16) and $FF, 2);
end;

function TXML2HTML.StyleToHTML(Name: string; Font: TFont): string;
begin
  result := Name + '{color: ' + Color2HTML(Font.Color) + ';';
  if Font.Name <> __DEFAULT_FONT_NAME then result := result + 'font-family: ' + Font.Name + ';';
  if Font.Size <> __DEFAULT_FONT_SIZE then result := result + 'font-size: ' + IntToStr(Font.Size) + 'px;';
  if fsBold in Font.Style then result := result + 'font-weight:bold;';
  if fsItalic in Font.Style then result := result + 'font-style:italic;';
  if fsUnderline in Font.Style then result := result + 'text-decoration:underline;';
  if fsStrikeOut in Font.Style then result := result + 'text-decoration:line-through;';
  result := result + ' }';
end;

function TXML2HTML.GetAsHTML: WideString;
var
  i, j: integer;
  LineInfo: TLineInfo;
  WordInfo: TWordInfo;
  HTML: TWideStringList;
  sHTML: WideString;
begin
  result := '';
  HTML := TWideStringList.Create;
  try
    if FProlog <> '' then HTML.Add(FProlog)
    else HTML.Add('<?xml version="1.0" encoding="ISO-8859-1" ?>');
    HTML.Add('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">');
    HTML.Add('<html xmlns="http://www.w3.org/1999/xhtml">');
    HTML.Add('<head>');
    HTML.Add('<title>XML as HTML</title>');

    HTML.Add('<style type="text/css">');
    HTML.Add('*{ margin: 0px;	padding: 0px;}');
    HTML.Add('html,body{background: #FFFFFF; }');
    HTML.Add('code{font-family: ' + __DEFAULT_FONT_NAME + ', sans-serif; font-size: ' +
              IntToStr(__DEFAULT_FONT_SIZE) + 'px; color: ' + Color2HTML(__DEFAULT_FONT_COLOR) + '; }');

    HTML.Add(StyleToHTML('#op', FFontOperator));
    HTML.Add(StyleToHTML('#att', FFontAttribute));
    HTML.Add(StyleToHTML('#qut', FFontQuote));
    HTML.Add(StyleToHTML('#val', FFontValue));
    HTML.Add(StyleToHTML('#com', FFontComment));
    HTML.Add(StyleToHTML('#txt', FFontText));
    HTML.Add(StyleToHTML('#cdata', FFontCDATA));
    HTML.Add(StyleToHTML('#proc', FFontProcessing));
    HTML.Add(StyleToHTML('#doc', FFontDoctype));
    HTML.Add(StyleToHTML('#tag', FFontTag));

    HTML.Add('</style>');
    HTML.Add('</head>');
    HTML.Add('<body>');
    HTML.Add('<code>');

    for i := 0 to FLines.Count - 1 do begin
      LineInfo := TLineInfo(FLines[i]);
      sHTML := '<span id="' + StyleToString(X2H_XML.stText) + '">' + GetHTMLBlockIndention(LineInfo.Level) + '</span>';
      for j := 0 to LineInfo.WordList.Count - 1 do begin
        WordInfo := TWordInfo(LineInfo.WordList[j]);
        sHTML := sHTML + '<span id="' + StyleToString(WordInfo.StyleType) + '">' + ReplaceHTMLChars(WordInfo.Text) + '</span>'
      end;
      HTML.Add(sHTML + '<br />');
    end;

    HTML.Add('</code>');
    HTML.Add('</body>');
    HTML.Add('</html>');

    result := HTML.Text;
  finally
    HTML.Free;
  end;
end;

end.


